#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>

typedef unsigned char byte;

#define TITLEBAR_HEIGHT 22

// the default value of iTerm
//#define TITLEBAR_HEIGHT 97

#define MAX_IMAGE 1000
typedef struct {
    NSImage* img;
    NSData* dat;
    NSBitmapImageRep* rep;
    byte* buf;
    int w, h;
    int bpp;
} IMG;
IMG imgs[MAX_IMAGE];

void clearImage(char* cmd, byte* dst, int dpitch) {
    int x, y, w, h;
    sscanf(cmd, "%d;%d;%d;%d", &x, &y, &w, &h);
    int i, j;
    for (j = y; j < y+h; j++) {
        for (i = x; i < x+w; i++) {
            dst[j*dpitch+i*4  ] = 0;
            dst[j*dpitch+i*4+1] = 0;
            dst[j*dpitch+i*4+2] = 0;
            dst[j*dpitch+i*4+3] = 0;
        }
    }
}

BOOL loadImage(int n, char* file) {
    NSImage* img = [[NSImage alloc] initByReferencingFile:
                                        [NSString stringWithCString: file]];
    if (!img) {
        /* error */
        return NO;
    }
    NSData* srcData = [img TIFFRepresentation];
    if (!srcData) {
        /* error */
        return NO;
    }
    NSBitmapImageRep* srcRep = [[NSBitmapImageRep alloc]
                                   initWithData: srcData];

    if (imgs[n].img) [imgs[n].img release];
    if (imgs[n].dat) [imgs[n].dat release];
    if (imgs[n].rep) [imgs[n].rep release];
    imgs[n].img = img;
    imgs[n].dat = srcData;
    imgs[n].rep = srcRep;
    imgs[n].buf = [srcRep bitmapData];
    imgs[n].w = [img size].width;
    imgs[n].h = [img size].height;
    imgs[n].bpp = [srcRep bitsPerPixel];

    return YES;
}

void drawImage(char* cmd, BOOL redraw, byte* dst, int dpitch) {
    int n, x, y, w, h, sx, sy, sw, sh;
    char file[256];
    sscanf(cmd, "%d;%d;%d;%d;%d;%d;%d;%d;%d;%s",
           &n, &x, &y, &w, &h, &sx, &sy, &sw, &sh, file);

    if (n < 0 || n >= MAX_IMAGE) {
        /* error */
        return;
    }
    if (redraw) {
        if (imgs[n].buf == NULL) {
            /* error? */
            return;
        }
    }
    else {
        if (!loadImage(n, file)) {
            return;
        }
    }

    int iw = imgs[n].w;
    int ibpp = imgs[n].bpp;
    if (ibpp != 24 && ibpp != 32) {
        /* error? */
        return;
    }
    int ispp = ibpp / 8;
    int i, j;
    byte* src = imgs[n].buf;

    for (j = 0; j < sh; j++) {
        for (i = 0; i < sw; i++) {
            dst[(y+j)*dpitch+(x+i)*4  ] = src[((sy+j)*iw+sx+i)*ispp];
            dst[(y+j)*dpitch+(x+i)*4+1] = src[((sy+j)*iw+sx+i)*ispp+1];
            dst[(y+j)*dpitch+(x+i)*4+2] = src[((sy+j)*iw+sx+i)*ispp+2];
            dst[(y+j)*dpitch+(x+i)*4+3] = 255;
//            dst[py*dpitch+px*4+3] = ispp == 3 ? 255 : src[(j*iw+i)*4+3];
        }
    }
}

typedef void* CGSValue;

int _CGSDefaultConnection();
void CGSGetOnScreenWindowCount(int cid, int wid, int* cnt);
void CGSGetOnScreenWindowList(int cid, int wid, int cnt, int* ids, int* acnt);
void CGSGetWindowProperty(int cid, int wid, CGSValue key, CGSValue* val);
CGSValue CGSCreateCStringNoCopy(char* str);
void CGSGetScreenRectForWindow(int cid, int wid, CGRect* rect);
char* CGSCStringValue(CGSValue str);

static int _focused_window(int* x, int* y, int* w, int *h) {
    int cid, cnt, *ids;
    int i;

    cid = _CGSDefaultConnection();
    CGSGetOnScreenWindowCount(cid, NO, &cnt);
    ids = (int*)malloc(sizeof(int)*cnt);
    CGSGetOnScreenWindowList(cid, NO, cnt, ids, &cnt);
    for (i = 0; i < cnt; i++) {
        CGSValue windowTitle;
        char* title;
        CGSGetWindowProperty(cid, ids[i],
                             CGSCreateCStringNoCopy("kCGSWindowTitle"),
                             &windowTitle);
        title = CGSCStringValue(windowTitle);
        if (title && *title) {
            CGRect rect;
            CGSGetScreenRectForWindow(cid, ids[i], &rect);
            *x = (int)rect.origin.x;
            *y = (int)rect.origin.y;
            *w = (int)rect.size.width;
            *h = (int)rect.size.height;

            NSRect scrrect = [[NSScreen mainScreen] frame];
            *y = scrrect.size.height - *y - *h;
            *h -= TITLEBAR_HEIGHT;

            free(ids);
            return 0;
        }
    }
    free(ids);
    return -1;
}

@interface W3MImg : NSView
{
    NSBitmapImageRep* bitmap;
    byte* pixels;
    NSImage* image;
}
- (void) prepare: (NSRect) r;
- (void) drawRect: (NSRect) r;
- (byte*) pixels;
- (int) bytesPerRow;
@end

@interface W3MImgMain : NSObject
- (void) applicationDidFinishLaunching: (NSNotification*) n;
@end

NSApplication* app;

@implementation W3MImg
- (byte*) pixels
{
    return pixels;
}

- (int) bytesPerRow
{
    return [bitmap bytesPerRow];
}

- (void) prepare: (NSRect) r
{
    bitmap = [[[NSBitmapImageRep alloc]
                  initWithBitmapDataPlanes: NULL
                  pixelsWide: r.size.width
                  pixelsHigh: r.size.height
                  bitsPerSample: 8
                  samplesPerPixel: 4
                  hasAlpha: YES
                  isPlanar: NO
                  colorSpaceName: NSDeviceRGBColorSpace
                  bytesPerRow: NO
                  bitsPerPixel: NO] autorelease];

    pixels = [bitmap bitmapData];

/*
    int x, y;
    byte* p = pixels;
    for (y = 0; y < r.size.height; y++) {
        for (x = 0; x < r.size.width; x++) {
            *p++ = (byte)y;
            *p++ = (byte)y;
            *p++ = (byte)y;
            *p++ = 200;
        }
    }
*/

    image = [[NSImage alloc] initWithSize: r.size];
    [image addRepresentation: bitmap];
}

- (void) drawRect: (NSRect) r
{
    [[NSColor clearColor] set];
    NSRectFill(r);

    [image compositeToPoint: NSMakePoint(0, 0)
           operation: NSCompositeSourceOver];
}
@end

@implementation W3MImgMain
- (void) applicationDidFinishLaunching: (NSNotification*) n
{
    int wx, wy, ww, wh;
    _focused_window(&wx, &wy, &ww, &wh);

    NSRect winrect = NSMakeRect(wx, wy, ww, wh);

    NSWindow* win = [[NSWindow alloc] initWithContentRect: winrect
                                      styleMask: NSBorderlessWindowMask
                                      backing: NSBackingStoreBuffered
                                      defer: 0];

    [win setOpaque: 0];

    W3MImg* v = [[W3MImg alloc] init];
    [v prepare: winrect];
    [win setContentView: v];

    [win makeKeyAndOrderFront: NULL];
    [win orderFrontRegardless];

    [win invalidateShadow];

    int done = NO;
    char buf[256];
    while (!done && fgets(buf, 255, stdin)) {
        int redraw = YES;

/*
        {
            FILE* a = fopen("w3mimg.log", "a");
            fputs(buf, a);
            fclose(a);
        }
*/

        switch (buf[0]) {
        case '0':
            redraw = NO;
        case '1':
            drawImage(buf+2, redraw, [v pixels], [v bytesPerRow]);
            break;
        case '2':
            done = YES;
            break;
        case '3':
            [v setNeedsDisplay: YES];
            [v display];
            [win makeKeyAndOrderFront: NULL];
            [win orderFrontRegardless];
            [win invalidateShadow];
            break;
        case '4':
            fprintf(stdout, "\n");
            fflush(stdout);
            break;
        case '5':
            break;
        case '6':
            clearImage(buf+2, [v pixels], [v bytesPerRow]);
            break;
        }
    }

    exit(0);
}
@end

int main(int argc, char* argv[]) {
    NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];

    if (argc > 1 && !strcmp(argv[1], "-test")) {
        int wx, wy, ww, wh;
        _focused_window(&wx, &wy, &ww, &wh);
        printf("%d %d\n", ww, wh);
        goto done;
    }
    else if (argc > 1 && !strcmp(argv[1], "-size")) {
        loadImage(0, argv[2]);
        fprintf(stdout, "%d %d\n", imgs[0].w, imgs[0].h);
        fflush(stdout);
        goto done;
    }

    W3MImgMain* ts = [[W3MImgMain alloc] init];
    app = [NSApplication sharedApplication];
    [app setDelegate: ts];
    [app run];

done:

    [pool release];
    return 0;
}